from typing import Any, Callable, Optional, Type, TypeVar

import torch
from torch import Tensor, nn
from pathlib import Path
from transformers import BioGptTokenizer
from transformers.models.biogpt.configuration_biogpt import BioGptConfig

from unicorn_baseline.vision_language.prism.biogpt_hf import BioGptForCausalLM


class BioGPT(nn.Module):
    def __init__(
        self,
        config: BioGptConfig,
        model_dir: Path,
        frozen_weights: bool = True,
        frozen_embeddings: bool = False,
        context_dim: int = 1280,
    ):
        super().__init__()

        self.context_dim = context_dim

        self.model = BioGptForCausalLM(config, x_attn_dim=context_dim)

        commit_hash = 'eb0d815e95434dc9e3b78f464e52b899bee7d923'
        tokenizer_weights = model_dir / 'biogpt'
        self.tokenizer = enforce_type(
            BioGptTokenizer,
            BioGptTokenizer.from_pretrained(tokenizer_weights, revision=commit_hash),
        )

        self.text_dim = enforce_type(
            torch.nn.Embedding, self.model.biogpt.embed_tokens
        ).embedding_dim

        # choose cls token from unused tokens
        # token 42383 : "madeupword0006</w>"
        # see https://huggingface.co/microsoft/biogpt/raw/main/vocab.json
        self.cls_token_id = 42383
        # init to random weight from the same distribution as original embeddings init
        enforce_type(Tensor, self.model.biogpt.embed_tokens.weight)[self.cls_token_id].data.normal_(
            mean=0.0, std=self.model.config.initializer_range
        )

        if frozen_weights:
            for name, param in self.model.named_parameters():
                if not any(c in name for c in ['embed_tokens', 'x_attn']):
                    param.requires_grad = False

        if frozen_embeddings:
            for name, param in self.model.named_parameters():
                if 'embed_tokens' in name:
                    param.requires_grad = False

    @property
    def pad_id(self) -> int:
        return enforce_type(int, self.tokenizer.pad_token_id)

    @property
    def bos_token_id(self) -> int:
        # biogpt uses EOS as BOS
        return enforce_type(int, self.tokenizer.eos_token_id)

    @property
    def eos_token_id(self) -> int:
        return enforce_type(int, self.tokenizer.eos_token_id)

    def tokenize(self, text: list[str]) -> Tensor:
        # add EOS token
        text = [t + self.tokenizer.eos_token for t in text]

        text_tokenised = self.tokenizer(
            text=text,
            add_special_tokens=True,
            padding=True,
            return_tensors='pt',
            return_token_type_ids=False,
            return_attention_mask=False,
        )

        # example:
        # "</s>Squamous mucosa indicative of reflux esophagitis. </s>"
        # [    2, 21477,  2626,  7265,     5,  4532, 13893,     4,     2]
        text_token_ids = enforce_type(Tensor, text_tokenised['input_ids'])
        return text_token_ids

    def untokenize(self, token_ids: Tensor) -> list[str]:
        return self.tokenizer.batch_decode(token_ids, skip_special_tokens=False)

    __call__: Callable[..., dict[str, Tensor]]

    def forward(
        self,
        input_ids: Tensor,
        key_value_states: Tensor,
        attention_mask: Optional[Tensor] = None,
        head_mask: Optional[Tensor] = None,
        past_key_values: Optional[tuple[tuple[Tensor]]] = None,
        use_cache: bool = False,
    ) -> dict[str, Tensor]:
        """
        Args:
            input_ids (`torch.LongTensor` of shape `(batch_size, sequence_length)`):
                Indices of input sequence tokens in the vocabulary.
                Call `self.tokenize` method to obtain token indices.

            key_value_states (`torch.FloatTensor` of shape `(batch_size, context_length, context_dim)`):
                Context for cross-attention modules in BioGPT; specifically, image latents from Perceiver.

            attention_mask (`torch.FloatTensor` of shape `(batch_size, sequence_length)`, *optional*):
                Mask to avoid performing attention on padding token indices. Mask values selected in `[0, 1]`:

                - 1 for tokens that are **not masked**,
                - 0 for tokens that are **masked**.

                If not provided, the mask will be computed using positions of pad token in `input_ids`.

            head_mask (`torch.FloatTensor` of shape `(num_heads,)` or `(num_layers, num_heads)`, *optional*):
                Mask to nullify selected heads of the self-attention modules. Mask values selected in `[0, 1]`:

                - 1 indicates the head is **not masked**,
                - 0 indicates the head is **masked**.

            past_key_values (`tuple(tuple(torch.FloatTensor))`, *optional*, returned when `use_cache=True` is passed or when `config.use_cache=True`):
                Tuple of `tuple(torch.FloatTensor)` of length `config.n_layers`, with each tuple having 2 tensors of shape
                `(batch_size, num_heads, sequence_length, embed_size_per_head)`) and 2 additional tensors of shape
                `(batch_size, num_heads, encoder_sequence_length, embed_size_per_head)`.

                Contains pre-computed hidden-states (key and values in the self-attention blocks and in the cross-attention
                blocks) that can be used (see `past_key_values` input) to speed up sequential decoding.

                If `past_key_values` are used, the user can optionally input only the last `decoder_input_ids` (those that
                don't have their past key value states given to this model) of shape `(batch_size, 1)` instead of all
                `decoder_input_ids` of shape `(batch_size, sequence_length)`. inputs_embeds (`torch.FloatTensor` of shape
                `(batch_size, sequence_length, hidden_size)`, *optional*): Optionally, instead of passing `input_ids` you
                can choose to directly pass an embedded representation. This is useful if you want more control over how to
                convert `input_ids` indices into associated vectors than the model's internal embedding lookup matrix.

            use_cache (`bool`, *optional*):
                If set to `True`, `past_key_values` key value states are returned and can be used to speed up decoding (see
                `past_key_values`).
        """
        batch_size = input_ids.shape[0]
        device = input_ids.device

        # append cls token id
        cls_token_ids = torch.full((batch_size, 1), self.cls_token_id, device=device)
        input_ids = torch.cat([input_ids, cls_token_ids], dim=-1)

        # attention mask
        if attention_mask is None:
            attention_mask = (input_ids != self.pad_id).to(torch.int8)

        output = self.model(
            input_ids=input_ids,
            key_value_states=key_value_states,
            attention_mask=attention_mask,
            head_mask=head_mask,
            past_key_values=past_key_values,
            use_cache=use_cache,
            output_hidden_states=True,
            xattn_collect_attn=None,
            return_dict=True,
        )

        # remove multimodal cls token
        logits = output['logits'][:, :-1]

        # take unimodal cls token from the last unimodal layer
        last_unimodal_layer = self.model.biogpt.first_x_attn_layer - 1
        # NOTE: `+ 1` accounts for the 0th entry in hidden_states being
        # not a layer output but text embeddings
        text_cls_embedding = output['hidden_states'][last_unimodal_layer + 1][:, -1]

        return {'logits': logits, 'text_embedding': text_cls_embedding}


T = TypeVar('T')


def enforce_type(t: Type[T], o: Any) -> T:
    if not isinstance(o, t):
        raise TypeError(f'{type(o)=} != t')
    return o



"""
Based on Perceiver architecture
"""

import warnings
from typing import Callable, Optional

import torch
import torch.nn.functional as F
from environs import Env
from torch import Tensor, nn
from torch.backends.cuda import sdp_kernel

from unicorn_baseline.vision_language.prism.configuring_prism import PerceiverConfig

env = Env()

PERCEIVER_MEM_EFF_ATTN: bool = env.bool('PERCEIVER_MEM_EFF_ATTN', default=False)
if PERCEIVER_MEM_EFF_ATTN:
    warnings.warn('Perceiver: using memory-efficient attention')

try:
    from xformers.ops import memory_efficient_attention  # type: ignore
except ImportError:
    if PERCEIVER_MEM_EFF_ATTN:
        raise Exception(
            'Memory efficient attention flag is set (PERCEIVER_MEM_EFF_ATTN) '
            'but xformers lib is not available.'
        )
    pass


# Attention modules


class CrossAttention(nn.Module):
    def __init__(
        self,
        *,
        query_dim: int,
        context_dim: int,
        head_dim: int,
        heads: int,
        return_attn: bool = False,
        c_norm: bool = True,
    ) -> None:
        super().__init__()

        self.query_dim = query_dim
        self.context_dim = context_dim

        self.head_dim = head_dim
        self.heads = heads

        self.scale = self.head_dim**-0.5

        self.inner_dim = self.head_dim * self.heads

        self.return_attn = return_attn

        self.x_norm = nn.LayerNorm(self.query_dim)
        self.c_norm = nn.LayerNorm(self.context_dim) if c_norm is True else None

        self.to_q = nn.Linear(self.query_dim, self.inner_dim, bias=False)

        self.to_kv = nn.Linear(self.context_dim, self.inner_dim * 2, bias=False)

        self.to_out = nn.Linear(self.inner_dim, self.query_dim, bias=False)

    def forward(
        self,
        x: Tensor,
        c: Tensor | None = None,
        kvt: tuple[Tensor, Tensor] | None = None,
        attn_mask: Tensor | None = None,
    ) -> tuple[Tensor, tuple[Tensor, Tensor], Tensor]:
        """
        Args:
            x: queries
            c: context
            kvt: key-value cache (instead of context)
            attn_mask: mask out part of context since contexts can vary in length

        Returns:
            processed output queries, KV-cache, attention weights
        """
        Bx, Nx, Dimx = x.shape

        x = self.x_norm(x)
        c = self.c_norm(c) if self.c_norm is not None else c

        q: Tensor = self.to_q(x)
        q = q.reshape(Bx, Nx, self.heads, self.head_dim)

        if c is not None and kvt is None:
            Bc, Nc, _ = c.shape
            kv: Tensor = self.to_kv(c)
            kv = kv.reshape(Bc, Nc, 2, self.heads, self.head_dim)
            k, v = kv.unbind(2)
            kvt = (k, v)
        elif kvt is not None and c is None:
            k, v = kvt
            Bc, Nc, _, _ = k.shape
            assert (Bc, Nc) == (v.shape[0], v.shape[1])
        else:
            raise Exception(f'XOR(c, kvt) but got: {type(c)} and {type(kvt)}.')

        if attn_mask is not None:
            attn_mask = attn_mask.reshape(Bc, 1, Nx, Nc).expand(-1, self.heads, -1, -1)

        if self.return_attn:
            warnings.warn('XATTN RETURNS ATTN SCORES, ONLY FOR EVAL!')
            q = q.permute(0, 2, 1, 3)
            k = k.permute(0, 2, 1, 3)
            v = v.permute(0, 2, 1, 3)
            q = q * self.scale
            sim = q @ k.transpose(-2, -1)
            if attn_mask is not None:
                sim = sim.masked_fill(~attn_mask, -torch.finfo(sim.dtype).max)
            attn = sim.softmax(dim=-1)
            a = attn @ v
            a = a.transpose(1, 2)

        elif PERCEIVER_MEM_EFF_ATTN:
            assert q.shape == (Bx, Nx, self.heads, Dimx // self.heads)
            assert k.shape == (Bx, Nc, self.heads, Dimx // self.heads)
            assert v.shape == (Bx, Nc, self.heads, Dimx // self.heads)
            if attn_mask is not None:
                attn_bias = torch.zeros_like(attn_mask, dtype=q.dtype, device=q.device)
                attn_bias = attn_bias.masked_fill(~attn_mask, -torch.finfo(q.dtype).max)
            else:
                attn_bias = None
            a = memory_efficient_attention(
                q,
                k,
                v,
                attn_bias=attn_bias,
                p=0.0,
                scale=None,
                op=None,
                output_dtype=None,
            )
            attn = torch.empty(0)

        else:
            q = q.permute(0, 2, 1, 3)
            k = k.permute(0, 2, 1, 3)
            v = v.permute(0, 2, 1, 3)
            with sdp_kernel(enable_flash=False, enable_math=True, enable_mem_efficient=False):
                a: Tensor = F.scaled_dot_product_attention(q, k, v, attn_mask=attn_mask)
                attn = torch.empty(0)
            a = a.transpose(1, 2)

        c = a.reshape(Bx, Nx, self.inner_dim)

        o = self.to_out(c)

        return o, kvt, attn


class MHSA(nn.Module):
    def __init__(self, *, dim: int, num_heads: int):
        super().__init__()

        self.norm = nn.LayerNorm(dim)

        self.mha = nn.MultiheadAttention(
            embed_dim=dim,
            num_heads=num_heads,
            dropout=0.0,
            bias=False,
            add_bias_kv=False,
            add_zero_attn=False,
            kdim=dim,
            vdim=dim,
            batch_first=True,
        )

    def forward(self, x: Tensor) -> Tensor:
        x = self.norm(x)

        with sdp_kernel(
            enable_flash=False, enable_math=True, enable_mem_efficient=PERCEIVER_MEM_EFF_ATTN
        ):
            x, _ = self.mha(
                x,
                x,
                x,
                key_padding_mask=None,
                need_weights=False,
                attn_mask=None,
                average_attn_weights=False,
                is_causal=False,
            )

        return x


# Feed forward


class GEGLU(nn.Module):
    def forward(self, x: Tensor):
        x, gates = x.chunk(2, dim=-1)
        return x * F.gelu(gates)


class FeedForward(nn.Module):
    def __init__(
        self,
        *,
        dim: int,
        mult: int = 1,
        dropout: float = 0.0,
        activation: str = 'geglu',
    ):
        super().__init__()

        self.norm = nn.LayerNorm(dim)

        extra_dim = 1

        if activation == 'geglu':
            actfn = GEGLU
            extra_dim = 2
        elif activation == 'gelu':
            actfn = nn.GELU
        else:
            raise Exception(f'{activation=} not supported.')

        self.net = nn.Sequential(
            nn.Linear(dim, dim * mult * extra_dim),
            actfn(),
            nn.Linear(dim * mult, dim),
            nn.Dropout(dropout),
        )

    def forward(self, x: Tensor):
        return self.net(self.norm(x))


class Perceiver(nn.Module):
    def __init__(
        self,
        *,
        latent_seq: int = 512,
        latent_dim: int = 768,
        context_dim: int = 2560,
        mhsa_heads: int = 8,
        perceiver_depth: int = 8,
        transformer_depth: int = 6,
        share_xattn_start_layer: int = 1,
        share_tf_start_layer: int = 0,
        xattn_heads: int = 1,
        mlp_mult: int = 1,
        mlp_activation: str = 'geglu',
    ):
        super().__init__()

        assert perceiver_depth > 0
        assert share_xattn_start_layer >= 0
        assert share_tf_start_layer >= 0

        self.share_xattn_start_layer = share_xattn_start_layer
        self.share_tf_start_layer = share_tf_start_layer
        self.latent_seq = latent_seq
        self.mhsa_heads = mhsa_heads

        latent_weights = torch.randn(latent_seq, latent_dim)
        self.latents = nn.Parameter(latent_weights)

        # inspired by lucidrains Perceiver repo

        get_xattn = lambda: nn.ModuleDict(
            {
                'xattn': CrossAttention(
                    query_dim=latent_dim,
                    context_dim=context_dim,
                    head_dim=latent_dim // xattn_heads,
                    heads=xattn_heads,
                    return_attn=False,
                    # norming large contexts explodes memory
                    c_norm=False,
                ),
                'ff': FeedForward(
                    dim=latent_dim,
                    mult=mlp_mult,
                    activation=mlp_activation,
                ),
            }
        )

        get_mhsa = lambda: nn.ModuleDict(
            {
                'mhsa': MHSA(
                    dim=latent_dim,
                    num_heads=mhsa_heads,
                ),
                'ff': FeedForward(
                    dim=latent_dim,
                    mult=mlp_mult,
                    activation=mlp_activation,
                ),
            }
        )

        get_transformer = lambda: nn.ModuleList([get_mhsa() for _ in range(transformer_depth)])

        layers = []
        for i in range(perceiver_depth):
            layer = nn.ModuleDict(
                {
                    'xattn': (
                        get_xattn() if i <= self.share_xattn_start_layer else layers[-1]['xattn']
                    ),
                    'tf': (
                        get_transformer() if i <= self.share_tf_start_layer else layers[-1]['tf']
                    ),
                }
            )
            layers.append(layer)

        self.layers = nn.ModuleList(layers)

    def forward(self, context: Tensor, attn_mask: Tensor) -> Tensor:
        """
        Args:
            context: input sequence that Perceiver processes with its latent queries.
                shape: (batch size, sequence length, feature dim)
            attn_mask: mask out padded area for each context example in the batch;
                necessary for batch size > 1 since contexts can have different sequence lengths.
                shape: (batch size, sequence length)

        Returns:
            processed latent queries (image latents)
        """
        B, N, _ = context.shape

        assert len(attn_mask.shape) == 2
        assert attn_mask.shape[:1] == context.shape[:1]

        attn_mask = attn_mask.reshape(B, 1, N).expand(-1, self.latent_seq, -1)

        x = self.latents.unsqueeze(0).expand(B, -1, -1)

        kvt = None
        for i, l in enumerate(self.layers):
            xattn = l['xattn']  # type: ignore
            if i <= self.share_xattn_start_layer:
                # pass context, return key-value cache
                xattn_out, kvt, _ = xattn['xattn'](x, c=context, kvt=None, attn_mask=attn_mask)
            else:
                # pass key-value cache since the previous layer shares W_k and W_v with this one
                xattn_out, kvt, _ = xattn['xattn'](x, c=None, kvt=kvt, attn_mask=attn_mask)

            x = xattn_out + x
            x = xattn['ff'](x) + x

            tf = l['tf']  # type: ignore
            for mhsa in tf:
                x = mhsa['mhsa'](x) + x
                x = mhsa['ff'](x) + x

        return x


class PerceiverResampler(nn.Module):
    def __init__(self, config: PerceiverConfig):
        super().__init__()

        self.query_dim = config.latent_dim
        self.query_seq = config.latent_seq

        latent_seq = 1 + config.latent_seq  # +1 for image embedding latent query

        self.cls_norm = nn.LayerNorm(config.latent_dim)
        self.img_norm = nn.LayerNorm(config.latent_dim)

        self.perceiver = Perceiver(
            latent_seq=latent_seq,
            latent_dim=config.latent_dim,
            context_dim=config.context_dim,
            mhsa_heads=config.mhsa_heads,
            perceiver_depth=config.perceiver_depth,
            transformer_depth=config.transformer_depth,
            share_xattn_start_layer=config.share_xattn_start_layer,
            share_tf_start_layer=config.share_tf_start_layer,
            xattn_heads=config.xattn_heads,
            mlp_mult=config.mlp_mult,
            mlp_activation=config.mlp_activation,
        )

    __call__: Callable[..., dict[str, Tensor]]

    def forward(
        self,
        tile_embeddings: Tensor,
        tile_mask: Optional[Tensor] = None,
    ) -> dict[str, Tensor]:
        """
        Args:
            tile_embeddings: sequences of tile embeddings per whole slide image
                shape: (batch size, sequence length, feature dim)
            attn_mask: mask out padded area for each sequence in the batch;
                necessary for batch size > 1 since images can have different sequence lengths.
                shape: (batch size, sequence length)

        Returns:
            dict with image embedding and image latents per sequence
        """
        batch_size = len(tile_embeddings)
        device = tile_embeddings.device

        if tile_mask is None:
            if batch_size > 1:
                raise Exception('tile pad mask must be provided with batch size>1.')

            tile_mask = torch.ones(
                tile_embeddings.shape[:2], device=device, dtype=tile_embeddings.dtype
            )

        x: Tensor = self.perceiver(context=tile_embeddings, attn_mask=tile_mask)

        cls_emb, img_emb = x[:, 0], x[:, 1:]

        assert cls_emb.shape == (len(tile_embeddings), self.query_dim)
        assert img_emb.shape == (len(tile_embeddings), self.query_seq, self.query_dim)

        cls_emb = self.cls_norm(cls_emb)
        img_emb = self.img_norm(img_emb)

        return {'image_embedding': cls_emb, 'image_latents': img_emb}



import argparse

import torch
from transformers import AutoModel

from .perceiver import PerceiverResampler

parser = argparse.ArgumentParser()
parser.add_argument('b', type=int, help='batch size', default=4)
parser.add_argument('n', type=int, help='number of tiles', default=100_000)

args = parser.parse_args()

print(args)

with torch.autocast('cuda', torch.float16), torch.inference_mode():
    prism = AutoModel.from_pretrained('paige-ai/Prism', trust_remote_code=True)

    net = prism.image_resampler.to('cuda')
    assert isinstance(net, PerceiverResampler), type(net)

    mem = 0
    for p in net.parameters():
        acc = 1
        for s in p.data.shape:
            acc *= s
        mem += acc

    print('number of parameters: ', round(mem / 1e6, 1), 'mn')

    print('init alloc:', torch.cuda.max_memory_allocated('cuda') / 1024**3, 'GB')
    print('init resrv:', torch.cuda.max_memory_reserved('cuda') / 1024**3, 'GB')

    batch_size = args.b
    seq_len = args.n
    e = torch.randn((batch_size, seq_len, 2560), device='cuda', dtype=torch.float16)
    m = torch.ones((batch_size, seq_len), device='cuda', dtype=torch.bool)

    print('data alloc:', torch.cuda.max_memory_allocated('cuda') / 1024**3, 'GB')
    print('data resrv:', torch.cuda.max_memory_reserved('cuda') / 1024**3, 'GB')

    net(tile_embeddings=e, tile_mask=m)

    print('forward alloc:', torch.cuda.max_memory_allocated('cuda') / 1024**3, 'GB')
    print('forward resrv:', torch.cuda.max_memory_reserved('cuda') / 1024**3, 'GB')



TASK_DEFINITIONS = {
    "Task012": {
        "Task": "Histopathology sample type",
        "Type": "Classification",
        "Description": (
            "This task involves extracting the origin of the material described in the pathology report. "
            "The output should classify the tissue origin into one of the following categories: lung, lymph node, "
            "bronchus, liver, brain, bone, or other. The origin of the tissue is generally mentioned at the beginning "
            "of the report as aard materiaal."
        ),
        "Parser_Format": {
            "single_label_multi_class_classification": {
                "type": "str",
                "description": (
                    "The origin of the material. One of 'lung', 'lymph node', 'bronchus', 'liver', "
                    "'brain', 'bone', or 'other'."
                ),
                "literals": [
                    "lung",
                    "lymph node",
                    "bronchus",
                    "liver",
                    "brain",
                    "bone",
                    "other",
                ],
            }
        },
    },
    "Task013": {
        "Task": "Pulmonary nodule presence",
        "Type": "Classification",
        "Description": (
            "This task requires analyzing the text of radiology reports to identify whether a pulmonary nodule is "
            "specifically mentioned. It is only relevant whether one is written literally in the text or not. "
            "You should not make inferences of the patient's health based on the report. The result should be a binary "
            "classification: 1.0 if a nodule is mentioned, and 0.0 if it is not."
        ),
        "Parser_Format": {
            "single_label_binary_classification": {
                "type": "float",
                "description": (
                    "The final classification. 1.0 if a pulmonary nodule is mentioned in the text, and 0.0 if it is not."
                ),
            }
        },
    },
    "Task014": {
        "Task": "Kidney Abnormality Presence",
        "Type": "Classification",
        "Description": (
            "This task involves determining whether a radiology report mentions any abnormalities related to the kidneys. "
            "Abnormalities include renal cell carcinoma, angiomyolipoma, cysts, kidney stones, conjoined kidneys, cases "
            "with partial or full nephrectomy, and several other rare abnormalities. The output should be a binary "
            "classification: 1.0 if a kidney abnormality is mentioned, and 0.0 if it is not."
        ),
        "Parser_Format": {
            "single_label_binary_classification": {
                "type": "float",
                "description": (
                    "The final classification. 1.0 if a kidney abnormality is mentioned in the text, and 0.0 if it is not."
                ),
            }
        },
    },
    "Task015": {
        "Task": "Hip Kellgren-Lawrence scoring",
        "Type": "Classification",
        "Description": (
            "This task involves classifying the Kellgren-Lawrence grade of osteoarthritis for both the left and right sides "
            "as described in the radiology report. The grades range from 0 to 4, with additional categories for 'not applicable (n)' "
            "and 'prosthesis (p)'. The output should provide a classification for each side."
        ),
        "Parser_Format": {
            "left": {
                "type": "str",
                "description": (
                    "The Kellgren-Lawrence grade of osteoarthritis for the left hip. An integer from 0 to 4, "
                    "or one of 'n' for not applicable or 'p' for prosthesis."
                ),
                "literals": ["0", "1", "2", "3", "4", "n", "p"],
            },
            "right": {
                "type": "str",
                "description": (
                    "The Kellgren-Lawrence grade of osteoarthritis for the right hip. An integer from 0 to 4, "
                    "or one of 'n' for not applicable or 'p' for prosthesis."
                ),
                "literals": ["0", "1", "2", "3", "4", "n", "p"],
            },
        },
    },
    "Task016": {
        "Task": "Colon Histopathology Diagnosis",
        "Type": "Classification",
        "Description": (
            "For the given numeral, predict whether the specimen was obtained from 1) biopsy (true) or polypectomy (false), "
            "and whether the pathologist rated the specimen as 2) hyperplastic polyps, 3) low-grade dysplasia, 4) high-grade dysplasia, "
            "5) cancer, 6) serrated polyps, or 7) non-informative. Give a true or false answer for each of the categories."
        ),
        "Parser_Format": {
            "biopsy": {
                "type": "bool",
                "description": "True if the specimen was obtained from a biopsy, false if from a polypectomy.",
            },
            "hyperplastic_polyps": {
                "type": "bool",
                "description": "True if rated as hyperplastic polyps.",
            },
            "low_grade_dysplasia": {
                "type": "bool",
                "description": "True if rated as low-grade dysplasia.",
            },
            "high_grade_dysplasia": {
                "type": "bool",
                "description": "True if rated as high-grade dysplasia.",
            },
            "cancer": {"type": "bool", "description": "True if rated as cancer."},
            "serrated_polyps": {
                "type": "bool",
                "description": "True if rated as serrated polyps.",
            },
            "non_informative": {
                "type": "bool",
                "description": "True if the specimen is non-informative.",
            },
        },
    },
    "Task017": {
        "Task": "Predicting lesion size measurements",
        "Type": "Regression",
        "Description": (
            "You will be provided with a radiology report and instructed to extract one of the following lesion size measurements, "
            "depending on the specific task stated at the beginning of each case: (1) PDAC size, (2) pulmonary nodule size, or (3) RECIST target lesion sizes. "
            "For PDAC or pulmonary nodule tasks, extract the lesion size in millimeters and assign it to lesion_1; set lesion_2 through lesion_5 to 0. "
            "For RECIST tasks, extract the size in millimeters for each described target lesion (up to 5), and assign them to lesion_1 through lesion_5 in order; "
            "set any remaining lesion values to 0 if fewer than 5 are described."
        ),
        "Parser_Format": {
            "lesion_1": {
                "type": "float",
                "description": (
                    "The estimated size of either the PDAC, the largest pulmonary nodule, or the first RECIST target lesion in mm."
                ),
            },
            "lesion_2": {
                "type": "float",
                "description": "The size of the second RECIST lesion in mm, or 0 if not described.",
            },
            "lesion_3": {
                "type": "float",
                "description": "The size of the third RECIST lesion in mm, or 0 if not described.",
            },
            "lesion_4": {
                "type": "float",
                "description": "The size of the fourth RECIST lesion in mm, or 0 if not described.",
            },
            "lesion_5": {
                "type": "float",
                "description": "The size of the fifth RECIST lesion in mm, or 0 if not described.",
            },
        },
    },
    "Task018": {
        "Task": "Predicting prostate volume, PSA, and PSA density",
        "Type": "Regression",
        "Description": (
            "This task involves one of three possible predictions, which will be specified at the beginning of each case: "
            "(1) predicting prostate volume in cubic centimeters, (2) predicting Prostate-Specific Antigen (PSA), or (3) predicting PSA density. "
            "If a range is given for prostate volume, calculate and use the average. PSA density is related to PSA and volume via the formula: "
            "volume = PSA / PSA density. Use this relationship if needed to compute missing values."
        ),
        "Parser_Format": {
            "single_label_regression": {
                "type": "float",
                "description": (
                    "The prostate volume in cubic centimeters, the PSA value, or the PSA density value depending on the case."
                ),
            }
        },
    },
    "Task019": {
        "Task": "Anonymization",
        "Type": "Single Label Named Entity Recognition",
        "Description": (
            "Identify and classify sequences of tokens in the given text that qualify as Personally Identifiable Information (PII). "
            "Replace each identified sequence with the corresponding predefined category tag. "
            "Categories include: <DATUM>, <PERSOON>, <RAPPORT_ID>, <PLAATS>, <PHINUMMER>, <STUDIE_NAAM>, "
            "<ACCREDITATIE_NUMMER>, <TIJD>, and <LEEFTIJD>. Use tags exactly as specified, and avoid tagging non-PII data like medical measurements."
        ),
        "Parser_Format": {
            "text": {
                "type": "str",
                "description": (
                    "The text with the identified PII entities replaced by their respective tags."
                ),
            }
        },
    },
}



from pathlib import Path

from unicorn_baseline.io import resolve_image_path, write_json_file
from unicorn_baseline.vision.pathology.main import extract_coordinates, save_coordinates
from unicorn_baseline.vision.pathology.wsi import FilterParams, TilingParams
from unicorn_baseline.vision_language.inference import generate_caption
from unicorn_baseline.vision_language.models import PRISM, Virchow


def get_file_path(file_location, extensions):
    for ext in extensions:
        potential_path = Path(f"{file_location}{ext}")
        if potential_path.exists():
            return potential_path
    return None


def save_output(caption, name):
    output_dict = [{"uid": name, "text": caption}]
    slug = "nlp-predictions-dataset"
    output_path = Path("/output")
    output_filename = output_path / f"{slug}.json"

    write_json_file(
        location=output_filename,
        content=output_dict,
    )

    print(f"Caption saved to {output_filename}")


def run_vision_language_task(*, input_information, model_dir):

    tissue_mask_path = None
    for input_socket in input_information:
        if input_socket["interface"]["kind"] == "Image":
            image_name = input_socket["image"]["name"]
            wsi_path = resolve_image_path(location=input_socket["input_location"])
        elif input_socket["interface"]["kind"] == "Segmentation":
            tissue_mask_path = resolve_image_path(location=input_socket["input_location"])

    num_workers = 4
    batch_size = 32
    mixed_precision = True
    max_number_of_tiles = 14000
    tiling_params = TilingParams(spacing=0.5, tolerance=0.07, tile_size=224, overlap=0.0, drop_holes=False, min_tissue_percentage=0.25, use_padding=True)
    filter_params = FilterParams(ref_tile_size=256, a_t=4, a_h=2, max_n_holes=8)

    # create output directories
    coordinates_dir = Path("/tmp/coordinates/")
    coordinates_dir.mkdir(parents=True, exist_ok=True)
    coordinates, _, level, resize_factor, tile_size_lv0, image_spacing, image_size = (
        extract_coordinates(
            wsi_path=wsi_path,
            tissue_mask_path=tissue_mask_path,
            tiling_params=tiling_params,
            filter_params=filter_params,
            max_number_of_tiles=max_number_of_tiles,
            num_workers=num_workers,
        )
    )

    save_coordinates(
        wsi_path=wsi_path,
        coordinates=coordinates,
        tile_level=level,
        tile_size=tiling_params.tile_size,
        resize_factor=resize_factor,
        tile_size_lv0=tile_size_lv0,
        target_spacing=tiling_params.spacing,
        save_dir=coordinates_dir,
    )

    tile_encoder = Virchow(model_dir=model_dir)
    prism = PRISM(model_dir=model_dir)

    caption = generate_caption(
        wsi_path,
        tile_encoder,
        prism,
        coordinates_dir,
        backend="asap",
        batch_size=batch_size,
        num_workers=num_workers,
        mixed_precision=mixed_precision,
    )

    caption = caption[0].replace("</s>", "").strip()

    save_output(caption, image_name)



import sys
from contextlib import nullcontext
from typing import Optional

import torch
import tqdm

from unicorn_baseline.vision.pathology.dataset import TileDataset, TileDatasetFromDisk


def extract_tile_features(model, dataloader, device, mixed_precision):
    """Extracts tile-level features using the given model."""
    features = []
    autocast_context = nullcontext()
    if mixed_precision:
        autocast_context = torch.autocast(device_type="cuda", dtype=torch.float16)
    with torch.inference_mode():
        with autocast_context:
            for _, image in tqdm.tqdm(
                dataloader,
                desc="Extracting tile-level features",
                unit=" tile",
                leave=True,
                file=sys.stdout,
            ):
                image = image.to(device, non_blocking=True)
                feature = model(image)
                features.append(feature.cpu())
    features = torch.cat(features, dim=0)
    return features.unsqueeze(0)


def get_slide_level_output(model, tile_features, device, mixed_precision):
    tile_features = tile_features.to(device)

    autocast_context = nullcontext()
    if mixed_precision:
        autocast_context = torch.autocast(device_type="cuda", dtype=torch.float16)

    with torch.inference_mode():
        with autocast_context:
            caption = model.forward_slide(
                tile_features,
            )
    return caption


def generate_caption(
    wsi_path,
    tile_encoder,
    slide_encoder,
    coordinates_dir,
    backend="asap",
    batch_size=1,
    num_workers=4,
    mixed_precision=False,
    load_tiles_from_disk=False,
    tile_format: Optional[str] = None,
):
    tile_encoder.eval().to(tile_encoder.device)
    slide_encoder.eval().to(slide_encoder.device)

    transforms = tile_encoder.get_transforms()

    if load_tiles_from_disk:
        dataset = TileDatasetFromDisk(
            wsi_path,
            tile_dir=coordinates_dir,
            tile_format=tile_format,
            transforms=transforms,
        )
    else:
        dataset = TileDataset(
            wsi_path, coordinates_dir=coordinates_dir, backend=backend, transforms=transforms
        )

    dataloader = torch.utils.data.DataLoader(
        dataset, batch_size=batch_size, num_workers=num_workers, pin_memory=True
    )

    tile_features = extract_tile_features(
        tile_encoder, dataloader, tile_encoder.device, mixed_precision
    )

    caption = get_slide_level_output(
        slide_encoder, tile_features, slide_encoder.device, mixed_precision
    )

    return caption



import torch
from pathlib import Path
from transformers import PreTrainedModel

from .vision_transformer import build_vision_tower
from .conch_v1_5 import build_conch
from .configuration_titan import TitanConfig


class Titan(PreTrainedModel):
    config_class = TitanConfig

    def __init__(self, config: TitanConfig, *model_args, **model_kwargs):
        super().__init__(config)

        self.vision_encoder = build_vision_tower(config.vision_config)
        self.conch_config = config.conch_config

    def return_conch(self, model_dir: Path):
        model, eval_transform = build_conch(self.conch_config, model_dir)
        return model, eval_transform

    def encode_slide_from_patch_features(self, patch_features: torch.Tensor, patch_coords: torch.Tensor, patch_size_lv0: int) -> torch.Tensor:
        '''
        encode whole-slide image using patch features
        Args:
            patch_features: torch.Tensor, shape (1, N, C)
            patch_coords: torch.Tensor, shape (1, N, 2)
            patch_size_lv0: int, patch size at level 0 (1024 if slide is 40x, 512 if slide is 20x)
        '''
        slide_embedding = self.vision_encoder(patch_features, patch_coords, patch_size_lv0, no_proj=True)
        return slide_embedding


from pathlib import Path

import numpy as np
import torch
import wholeslidedata as wsd
from PIL import Image


class TileDataset(torch.utils.data.Dataset):
    def __init__(self, wsi_path, coordinates_dir, backend, transforms=None):
        self.path = wsi_path
        self.backend = backend
        self.name = wsi_path.stem.replace(" ", "_")
        self.load_coordinates(coordinates_dir)
        self.transforms = transforms

    def set_coordinates(self, coordinates):
        self.coordinates = (np.array([coordinates["x"], coordinates["y"]]).T).astype(
            int
        )

    def load_coordinates(self, coordinates_dir):
        coordinates_file = coordinates_dir / f"{self.name}.npy"
        coordinates = np.load(coordinates_file, allow_pickle=True)
        self.x = coordinates["x"]
        self.y = coordinates["y"]
        self.tile_size_resized = coordinates["tile_size_resized"]
        self.tile_level = coordinates["tile_level"]
        self.resize_factor = coordinates["resize_factor"]
        self.tile_size_lv0 = coordinates["tile_size_lv0"][0]
        self.set_coordinates(coordinates)

    def __len__(self):
        return len(self.coordinates)

    def __getitem__(self, idx):
        wsi = wsd.WholeSlideImage(self.path, backend=self.backend)
        tile_level = self.tile_level[idx]
        tile_spacing = wsi.spacings[tile_level]
        tile_arr = wsi.get_patch(
            self.x[idx],
            self.y[idx],
            self.tile_size_resized[idx],
            self.tile_size_resized[idx],
            spacing=tile_spacing,
            center=False,
        )
        tile = Image.fromarray(tile_arr).convert("RGB")
        if self.resize_factor[idx] != 1:
            tile_size = int(self.tile_size_resized[idx] / self.resize_factor[idx])
            tile = tile.resize((tile_size, tile_size))
        if self.transforms:
            tile = self.transforms(tile)
        return idx, tile


class TileDatasetFromDisk(torch.utils.data.Dataset):
    def __init__(self, wsi_path, tile_dir, tile_format, transforms=None):
        self.name = wsi_path.stem.replace(" ", "_")
        self.tiles = sorted(
            [x for x in Path(tile_dir, self.name).glob(f"*.{tile_format}")]
        )
        self.transforms = transforms

    def __len__(self):
        return len(self.tiles)

    def __getitem__(self, idx):
        pil_tile = Image.open(self.tiles[idx])
        if self.transforms is not None:
            img = self.transforms(pil_tile)
        else:
            img = pil_tile
        return idx, img



from pathlib import Path
from typing import Any

from transformers import PretrainedConfig
from transformers.models.biogpt.configuration_biogpt import BioGptConfig


class PerceiverConfig(PretrainedConfig):
    model_type = "perceiver"

    def __init__(
        self,
        latent_seq: int = 512,
        latent_dim: int = 1280,
        context_dim: int = 2560,
        mhsa_heads: int = 8,
        perceiver_depth: int = 8,
        transformer_depth: int = 6,
        share_xattn_start_layer: int = 1,
        share_tf_start_layer: int = 0,
        xattn_heads: int = 1,
        mlp_mult: int = 1,
        mlp_activation: str = 'geglu',
        **kwargs: Any,
    ):
        self.latent_seq = latent_seq
        self.latent_dim = latent_dim
        self.context_dim = context_dim
        self.mhsa_heads = mhsa_heads
        self.perceiver_depth = perceiver_depth
        self.transformer_depth = transformer_depth
        self.share_xattn_start_layer = share_xattn_start_layer
        self.share_tf_start_layer = share_tf_start_layer
        self.xattn_heads = xattn_heads
        self.mlp_mult = mlp_mult
        self.mlp_activation = mlp_activation

        super().__init__(**kwargs)


class PrismConfig(PretrainedConfig):
    model_type = "prism"
    is_composition = True

    def __init__(
        self,
        biogpt_config: BioGptConfig,
        perceiver_config: PerceiverConfig,
        model_dir: Path,
        biogpt_context_dim: int = 1280,
        biogpt_frozen_weights: bool = True,
        biogpt_frozen_embeddings: bool = False,
        dim_latents: int = 5120,
        **kwargs: Any,
    ):
        if isinstance(biogpt_config, dict):
            self.biogpt_config = BioGptConfig(**biogpt_config)
        else:
            self.biogpt_config = biogpt_config

        if isinstance(perceiver_config, dict):
            self.perceiver_config = PerceiverConfig(**perceiver_config)
        else:
            self.perceiver_config = perceiver_config

        self.model_dir = model_dir
        self.biogpt_context_dim = biogpt_context_dim
        self.biogpt_frozen_weights = biogpt_frozen_weights
        self.biogpt_frozen_embeddings = biogpt_frozen_embeddings

        self.dim_latents = dim_latents

        super().__init__(**kwargs)



import multiprocessing as mp
import os
import random
import sys
from pathlib import Path
from typing import Any

import numpy as np
import tqdm
import wholeslidedata as wsd
from PIL import Image

from unicorn_baseline.io import resolve_image_path, write_json_file
from unicorn_baseline.vision.pathology.feature_extraction import extract_features
from unicorn_baseline.vision.pathology.models import TITAN
from unicorn_baseline.vision.pathology.wsi import TilingParams, FilterParams, WholeSlideImage


def extract_coordinates(
    *,
    wsi_path: Path,
    tissue_mask_path: Path,
    tiling_params: TilingParams,
    filter_params: FilterParams,
    max_number_of_tiles: int | None = None,
    seed: int = 0,
    num_workers: int = 1,
):
    """
    Extracts tile coordinates from a whole slide image (wsi) based on the given parameters.

    Args:
        wsi_path (str): File path to the wsi.
        tissue_mask_path (str): File path to the tissue mask associated with the wsi.
        tiling_params (TilingParams): Parameters for tiling the wsi.
        filter_params (FilterParams): Parameters for filtering the tiles.
        max_number_of_tiles (int, optional): Maximum number of tiles to keep. If None, all tiles are kept. Defaults to None.
        num_workers (int, optional): Number of workers to use for parallel processing. Defaults to 1.

    Returns:
        tuple: A tuple containing:
            - sorted_coordinates (list): List of tile coordinates sorted by tissue percentage.
            - sorted_tissue_percentages (list): List of tissue percentages corresponding to the sorted coordinates.
            - tile_level (int): Tile level used for extraction.
            - resize_factor (float): Resize factor applied to the tiles.
            - tile_size_lv0 (int): Tile size at level 0 in the wsi.
    """
    wsi = WholeSlideImage(wsi_path, tissue_mask_path)
    (
        coordinates,
        tissue_percentages,
        tile_level,
        resize_factor,
        tile_size_lv0,
    ) = wsi.get_tile_coordinates(
        tiling_params=tiling_params,
        filter_params=filter_params,
        num_workers=num_workers,
    )

    image_spacings = wsi.spacings[0]
    required_level, _ = wsi.get_best_level_for_spacing(tiling_params.spacing, tiling_params.tolerance)
    image_size = wsi.level_dimensions[required_level]

    # sort coordinates by tissue percentage
    sorted_coordinates, sorted_tissue_percentages = select_coordinates_with_tissue(
        coordinates=coordinates, tissue_percentages=tissue_percentages, max_number_of_tiles=max_number_of_tiles, seed=seed
    )

    return (
        sorted_coordinates,
        sorted_tissue_percentages,
        tile_level,
        resize_factor,
        tile_size_lv0,
        image_spacings,
        image_size,
    )


def save_coordinates(
    *,
    wsi_path: Path,
    coordinates,
    tile_level,
    tile_size,
    resize_factor,
    tile_size_lv0,
    target_spacing,
    save_dir: str,
):
    """
    Saves tile coordinates and associated metadata into a .npy file.

    Args:
        wsi_path (Path): File path to the whole slide image (wsi).
        coordinates (list of tuples): List of (x, y) coordinates of tiles, defined with respect to level 0.
        tile_level (int): Level of the image pyramid at which tiles have been extracted.
        tile_size (int): Desired tile size.
        resize_factor (float): Factor by which the tile size must be resized.
        tile_size_lv0 (int): Size of the tile at level 0.
        target_spacing (float): Target spacing for the tiles.
        save_dir (str): Directory where the output .npy file will be saved.

    Returns:
        Path: Path to the saved .npy file containing tile coordinates and associated metadata.

    Notes:
        - The output file is saved with the same name as the wsi file stem.
        - The metadata includes the resized tile size, tile level, resize factor, tile size at level 0,
          and target spacing for each tile.
    """
    wsi_name = wsi_path.stem
    output_path = Path(save_dir, f"{wsi_name}.npy")
    x = [c[0] for c in coordinates]  # defined w.r.t level 0
    y = [c[1] for c in coordinates]  # defined w.r.t level 0
    ntile = len(x)
    tile_size_resized = int(tile_size * resize_factor)

    dtype = [
        ("x", int),
        ("y", int),
        ("tile_size_resized", int),
        ("tile_level", int),
        ("resize_factor", float),
        ("tile_size_lv0", int),
        ("target_spacing", float),
    ]
    data = np.zeros(ntile, dtype=dtype)
    for i in range(ntile):
        data[i] = (
            x[i],
            y[i],
            tile_size_resized,
            tile_level,
            resize_factor,
            tile_size_lv0,
            target_spacing,
        )

    data_arr = np.array(data)
    np.save(output_path, data_arr)
    return output_path


def select_coordinates_with_tissue(
    *,
    coordinates: list[tuple],
    tissue_percentages: list[float],
    max_number_of_tiles: int | None = None,
    seed: int = 0,
) -> tuple[list[tuple], list[float]]:
    """
    Select coordinates and their corresponding tissue percentages based on a
    maximum number of tiles. If more than the maximum number of tiles are found
    with full tissue content (100%), a random selection of tiles is made.
    Otherwise, the tiles are sorted by tissue percentage and the top tiles are
    selected.

    Args:
        coordinates (list of tuple): A list of tuples representing coordinates,
            where each tuple contains two integers (x, y).
        tissue_percentages (list of float): A list of tissue percentages
            corresponding to the tile for each coordinate.
        max_number_of_tiles (int | None): The maximum number of tiles to select.
            If None, all tiles above the minimum tissue percentage will be selected.

    Returns:
        tuple: A tuple containing two lists:
            - sorted_coordinates (list of tuple): The coordinates sorted based
              on the tissue percentages.
            - sorted_tissue_percentages (list of float): The tissue percentages
              corresponding to the sorted coordinates.
    """

    # separate perfect tissue tiles
    perfect = [(coord, perc) for coord, perc in zip(coordinates, tissue_percentages) if perc == 1.0]
    if max_number_of_tiles is not None and len(perfect) > max_number_of_tiles:
        rng = random.Random(seed)
        selected = rng.sample(perfect, max_number_of_tiles)
    else:
        # Sort by descending tissue percentage and take top N if needed
        all = [(coord, perc) for coord, perc in zip(coordinates, tissue_percentages)]
        all.sort(key=lambda x: x[1], reverse=True)
        selected = all[:max_number_of_tiles] if max_number_of_tiles is not None else all

    selected_coordinates, selected_percentages = zip(*selected)
    return list(selected_coordinates), list(selected_percentages)


def save_tile(
    *,
    x: int,
    y: int,
    wsi_path: Path,
    spacing: float,
    tile_size: int,
    resize_factor: int | float,
    save_dir: Path,
    tile_format: str,
    backend: str = "asap",
):
    tile_size_resized = int(tile_size * resize_factor)
    wsi = wsd.WholeSlideImage(wsi_path, backend=backend)
    tile_arr = wsi.get_patch(
        x, y, tile_size_resized, tile_size_resized, spacing=spacing, center=False
    )
    tile = Image.fromarray(tile_arr).convert("RGB")
    if resize_factor != 1:
        tile = tile.resize((tile_size, tile_size))
    tile_fp = save_dir / f"{int(x)}_{int(y)}.{tile_format}"
    tile.save(tile_fp)
    return tile_fp


def save_tile_mp(args):
    coord, wsi_path, spacing, tile_size, resize_factor, tile_dir, tile_format = args
    x, y = coord
    return save_tile(
        x=x, y=y, wsi_path=wsi_path, spacing=spacing, tile_size=tile_size, resize_factor=resize_factor, save_dir=tile_dir, tile_format=tile_format
    )


def save_tiles(
    *,
    wsi_path: Path,
    coordinates: list[tuple[int, int]],
    tile_level: int,
    tile_size: int,
    resize_factor: float,
    save_dir: Path,
    tile_format: str,
    backend: str = "asap",
    num_workers: int = 1,
):
    wsi_name = wsi_path.stem
    wsi = wsd.WholeSlideImage(wsi_path, backend=backend)
    tile_spacing = wsi.spacings[tile_level]
    tile_dir = save_dir / wsi_name
    tile_dir.mkdir(parents=True, exist_ok=True)
    iterable = [
        (coord, wsi_path, tile_spacing, tile_size, resize_factor, tile_dir, tile_format)
        for coord in coordinates
    ]
    with mp.Pool(num_workers) as pool:
        for _ in tqdm.tqdm(
            pool.imap_unordered(save_tile_mp, iterable),
            desc=f"Saving tiles for {wsi_path.stem}",
            unit=" tile",
            total=len(iterable),
            leave=True,
            file=sys.stdout,
        ):
            pass


def save_feature_to_json(
    *,
    feature,
    task_type,
    title,
    coordinates=None,
    tile_size=None,
    spacing=None,
    image_size=None,
    image_spacing=None,
    image_origin=None,
    image_direction=None,
):
    """
    Saves the extracted feature vector to a JSON file in the required format.
    """
    if task_type in ["classification", "regression"]:
        output_dict = [{"title": title, "features": feature}]
        output_path = Path("/output")
        output_filename = output_path / "image-neural-representation.json"

    else:
        print("Spacing: ", spacing)
        if image_origin is None:
            image_origin = [0.0] * len(image_size)
        if image_direction is None:
            image_direction = np.identity(len(image_size)).flatten().tolist()

        output_dict = [
            {
                "title": title,  # Use WSI filename as title
                "patches": [
                    {
                        "coordinates": [int(coord[0]), int(coord[1])],
                        "features": feat.cpu().tolist(),
                    }
                    for coord, feat in zip(coordinates, feature)
                ],
                "meta": {
                    "patch-size": tile_size,
                    "patch-spacing": [spacing, spacing],
                    "image-size": image_size,
                    "image-origin": image_origin,
                    "image-spacing": [image_spacing, image_spacing],
                    "image-direction": image_direction,
                },
            }
        ]

        output_path = Path("/output")
        output_filename = output_path / "patch-neural-representation.json"

    write_json_file(
        location=output_filename,
        content=output_dict,
    )

    print(f"Feature vector saved to {output_filename}")


def run_pathology_vision_task(
    *,
    task_name: str,
    task_type: str,
    input_information: dict[str, Any],
    model_dir: Path,
):
    tissue_mask_path = None
    for input_socket in input_information:
        if input_socket["interface"]["kind"] == "Image":
            image_title = input_socket["image"]["pk"]
            wsi_path = resolve_image_path(location=input_socket["input_location"])
        elif input_socket["interface"]["kind"] == "Segmentation":
            tissue_mask_path = resolve_image_path(location=input_socket["input_location"])

    batch_size = 32
    use_mixed_precision = True
    save_tiles_to_disk = False
    tile_format = "jpg"
    max_number_of_tiles = 14000

    num_workers = min(mp.cpu_count(), 8)
    if "SLURM_JOB_CPUS_PER_NODE" in os.environ:
        num_workers = min(num_workers, int(os.environ["SLURM_JOB_CPUS_PER_NODE"]))

    # coonfigurations for tile extraction based on tasks
    clf_config = {
        "tiling_params": TilingParams(spacing=0.5, tolerance=0.07, tile_size=512, overlap=0.0, drop_holes=False, min_tissue_percentage=0.25, use_padding=True),
        "filter_params": FilterParams(ref_tile_size=256, a_t=4, a_h=2, max_n_holes=8),
    }

    detection_config = {
            "tiling_params": TilingParams(spacing=0.5, tolerance=0.07, tile_size=224, overlap=0, drop_holes=False, min_tissue_percentage=0.1, use_padding=True),
            "filter_params": FilterParams(ref_tile_size=64, a_t=1, a_h=1, max_n_holes=8),
        }

    if task_name == 'Task08_detecting_mitotic_figures_in_breast_cancer_wsis':
        detection_config = {
            "tiling_params": TilingParams(spacing=0.25, tolerance=0.07, tile_size=448, overlap=0, drop_holes=False, min_tissue_percentage=0.1, use_padding=True),
            "filter_params": FilterParams(ref_tile_size=64, a_t=1, a_h=1, max_n_holes=8),
        }

    segmentation_config = {
        "tiling_params": TilingParams(spacing=0.5, tolerance=0.07, tile_size=224, overlap=0, drop_holes=False, min_tissue_percentage=0.1, use_padding=True),
        "filter_params": FilterParams(ref_tile_size=64, a_t=1, a_h=1, max_n_holes=8),
    }

    task_configs = {
        "classification": clf_config,
        "regression": clf_config,
        "detection": detection_config,
        "segmentation": segmentation_config
    }

    config = task_configs[task_type]

    # create output directories
    coordinates_dir = Path("/tmp/coordinates/")
    coordinates_dir.mkdir(parents=True, exist_ok=True)

    # Extract tile coordinates
    coordinates, _, level, resize_factor, tile_size_lv0, image_spacing, image_size = (
        extract_coordinates(
            wsi_path=wsi_path,
            tissue_mask_path=tissue_mask_path,
            tiling_params=config["tiling_params"],
            filter_params=config["filter_params"],
            max_number_of_tiles=max_number_of_tiles,
            num_workers=num_workers,
        )
    )

    save_coordinates(
        wsi_path=wsi_path,
        coordinates=coordinates,
        tile_level=level,
        tile_size=config["tiling_params"].tile_size,
        resize_factor=resize_factor,
        tile_size_lv0=tile_size_lv0,
        target_spacing=config["tiling_params"].spacing,
        save_dir=coordinates_dir,
    )

    tile_dir = None
    if save_tiles_to_disk:
        tile_dir = Path("/tmp/tiles/")
        tile_dir.mkdir(parents=True, exist_ok=True)
        save_tiles(
            wsi_path=wsi_path,
            coordinates=coordinates,
            tile_level=level,
            tile_size=config["tiling_params"].tile_size,
            resize_factor=resize_factor,
            save_dir=tile_dir,
            tile_format=tile_format,
            num_workers=num_workers,
        )

    print("=+=" * 10)
    feature_extractor = TITAN(model_dir)

    # Extract tile or slide features
    feature = extract_features(
        wsi_path=wsi_path,
        model=feature_extractor,
        coordinates_dir=coordinates_dir,
        task_type=task_type,
        backend="asap",
        batch_size=batch_size,
        num_workers=num_workers,
        use_mixed_precision=use_mixed_precision,
        load_tiles_from_disk=save_tiles_to_disk,
        tile_dir=tile_dir,
        tile_format=tile_format,
    )

    if task_type in ["classification", "regression"]:
        save_feature_to_json(feature=feature, task_type=task_type, title=image_title)
    elif task_type in ["detection", "segmentation"]:
        tile_size = [config["tiling_params"].tile_size, config["tiling_params"].tile_size, 3]

        save_feature_to_json(
            feature=feature,
            task_type=task_type,
            title=image_title,
            coordinates=coordinates,
            tile_size=tile_size,
            spacing=config["tiling_params"].spacing,
            image_size=image_size,
            image_spacing=image_spacing,
        )
